// Copyright (c) 2022 PaddlePaddle Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ultra_infer/function/cast.h"
#include "ultra_infer/function/clip.h"
#include "ultra_infer/function/concat.h"
#include "ultra_infer/function/cumprod.h"
#include "ultra_infer/function/elementwise.h"
#include "ultra_infer/function/full.h"
#include "ultra_infer/function/gather_scatter_along_axis.h"
#include "ultra_infer/function/gaussian_random.h"
#include "ultra_infer/function/isfinite.h"
#include "ultra_infer/function/linspace.h"
#include "ultra_infer/function/math.h"
#include "ultra_infer/function/pad.h"
#include "ultra_infer/function/quantile.h"
#include "ultra_infer/function/reduce.h"
#include "ultra_infer/function/slice.h"
#include "ultra_infer/function/softmax.h"
#include "ultra_infer/function/sort.h"
#include "ultra_infer/function/split.h"
#include "ultra_infer/function/tile.h"
#include "ultra_infer/function/transpose.h"
